# -*- coding: utf-8 -*-
"""Module to print neatly formatted tables and grids to the terminal."""

import enum
from typing import (
    Optional,
    List,
    Union,
    Iterator,
    Sequence,
    Any,
    Tuple,
    Callable,
    TYPE_CHECKING,
)

import click

if TYPE_CHECKING:
    from datetime import datetime


# ==== enums ===========================================================================


class Align(enum.Enum):
    """Text alignment in column"""

    Left = 0
    Right = 1


class Elide(enum.Enum):
    """Elide directives"""

    Leading = 0
    Center = 1
    Trailing = 2


class Prefix(enum.Enum):
    """Prefix for command line output"""

    Info = 0
    Ok = 1
    Warn = 2
    NONE = 3


# ==== text adjustment helpers =========================================================


def elide(
    text: str, width: int, placeholder: str = "...", elide: Elide = Elide.Trailing
) -> str:
    """
    Elides a string to fit into the given width.

    :param text: Text to truncate.
    :param width: Target width.
    :param placeholder: Placeholder string to indicate truncated text.
    :param elide: Which part to truncate.
    :returns: Truncated text.
    """

    if len(text) <= width:
        return text

    available = width - len(placeholder)

    if elide is Elide.Trailing:
        return text[:available] + placeholder
    elif elide is Elide.Leading:
        return placeholder + text[-available:]
    else:
        half_available = available // 2
        return text[:half_available] + placeholder + text[-half_available:]


def adjust(text: str, width: int, align: Align = Align.Left) -> str:
    """
    Pads a string with spaces up the desired width. Preserves ANSI color codes without
    counting them towards the width.

    This function is similar to ``str.ljust`` and ``str.rjust``.

    :param text: Initial string.
    :param width: Target width. If smaller than the given text, nothing is done.
    :param align: Side to align the padded string: to the left or to the right.
    """

    needed = width - len(click.unstyle(text))

    if needed > 0:
        if align == Align.Left:
            return text + " " * needed
        else:
            return " " * needed + text
    else:
        return text


# ==== printing structured data to console =============================================


class Field:
    """Base class to represent a field in a table."""

    @property
    def display_width(self) -> int:
        """
        The requested total width of the content in characters when not wrapped or
        shortened in any way.
        """
        raise NotImplementedError()

    def format(self, width: int) -> List[str]:
        """
        Returns the field content formatted to fit the requested width.

        :param width: Width to fit.
        :returns: Shortened or wrapped string.
        """
        raise NotImplementedError()


class TextField(Field):
    """
    A text field for a table.

    :param text: Text to represent.
    :param align: Text alignment: right or left.
    :param wraps: Whether to wrap the text instead of truncating it to fit into a
        requested width.
    :elide: Truncation strategy: trailing, center or leading.
    :param style: Styling passed on to :meth:`click.style` when styling the text.
    """

    def __init__(
        self,
        text: str,
        align: Align = Align.Left,
        wraps: bool = False,
        elide: Elide = Elide.Trailing,
        **style,
    ) -> None:
        self.text = text
        self.align = align
        self.wraps = wraps
        self.elide = elide
        self.style = style

    @property
    def display_width(self) -> int:
        return len(self.text)

    def format(self, width: int) -> List[str]:

        import textwrap

        if self.wraps:
            lines = textwrap.wrap(self.text, width=width)
        else:
            lines = [elide(self.text, width, elide=self.elide)]

        # apply style first, adjust later

        if self.style:
            lines = [click.style(line, **self.style) for line in lines]

        return [adjust(line, width, self.align) for line in lines]

    def __repr__(self):
        return f"<{self.__class__.__name__}('{self.text}')>"


class DateField(Field):
    """
    A datetime field for a table. The formatting of the datetime will be adjusted
    depending on the available width. Does not currently support localisation.

    :param dt: Datetime to represent.
    :param style: Styling passed on to :meth:`click.style` when styling the text.
    """

    def __init__(self, dt: "datetime", **style) -> None:
        self.dt = dt
        self.style = style

    @property
    def display_width(self) -> int:
        return 20

    def format(self, width: int) -> List[str]:
        if width >= 20:
            string = self.dt.strftime("%d %b %Y at %H:%M")
        elif width >= 17:
            string = self.dt.strftime("%x, %H:%M")
        else:
            string = self.dt.strftime("%x")

        if self.style:
            return [click.style(string, **self.style)]
        else:
            return [string]

    def __repr__(self):
        return f"<{self.__class__.__name__}('{self.format(17)}')>"


class Column:
    """
    A table column.

    :param title: Column title.
    :param fields: Fields in the table. Any sequence of objects can be given and will be
        converted to :class:`Field` instances as appropriate.
    :param align: How to align text inside the column. Will only be used for
        :class:`TextField`s.
    :param wraps: Whether to wrap fields to fit into the column width instead of
        truncating them. Will only be used for :class:`TextField`s.
    :param elide: How to elide text which is too wide for a column. Will only be used
        for :class:`TextField`s.
    """

    fields: List[Field]

    def __init__(
        self,
        title: Optional[str],
        fields: Sequence = (),
        align: Align = Align.Left,
        wraps: bool = False,
        elide: Elide = Elide.Trailing,
    ) -> None:
        self.title = TextField(title, align=align, bold=True) if title else None
        self.align = align
        self.wraps = wraps
        self.elide = elide

        self.fields = []

        for field in fields:
            self.fields.append(self._to_field(field))

    @property
    def display_width(self) -> int:
        if self.title:
            all_fields = self.fields + [self.title]
        else:
            all_fields = self.fields
        return max(field.display_width for field in all_fields)

    @property
    def has_title(self):
        return self.title is not None

    def append(self, field: Any) -> None:
        self.fields.append(self._to_field(field))

    def insert(self, index: int, field: Any) -> None:
        self.fields.insert(index, self._to_field(field))

    def __getitem__(self, item: int) -> Field:
        return self.fields[item]

    def __setitem__(self, key: int, value: Any) -> None:
        self.fields[key] = self._to_field(value)

    def __iter__(self) -> Iterator[Field]:
        return iter(self.fields)

    def __len__(self) -> int:
        return len(self.fields)

    def _to_field(self, field: Any) -> Field:

        from datetime import datetime

        if isinstance(field, Field):
            return field
        elif isinstance(field, datetime):
            return DateField(field)
        else:
            return TextField(
                str(field), align=self.align, wraps=self.wraps, elide=self.elide
            )

    def __repr__(self):
        title = self.title.text if self.title else "untitled"
        return f"<{self.__class__.__name__}(title='{title}')>"


class Table:
    """
    A table which can be printed to stdout.

    :param columns: Table columns. Can be a list of :class:`Column` instances or table
        titles.
    :param padding: Padding between columns.
    """

    columns: List[Column]

    def __init__(self, columns: List[Union[Column, str]], padding: int = 2) -> None:
        self.columns = []
        self.padding = padding

        for col in columns:
            if isinstance(col, Column):
                self.columns.append(col)
            else:
                self.columns.append(Column(col))

    @property
    def ncols(self) -> int:
        """The number of columns"""
        return len(self.columns)

    @property
    def nrows(self) -> int:
        """The number of rows"""
        return max(len(col) for col in self.columns)

    def append(self, row: Sequence) -> None:
        """
        Appends a new row to the table.

        :param row: List of fields to append to each column. Length must match the
            number of columns.
        """

        if len(row) != self.ncols:
            raise ValueError(f"Got {len(row)} fields but have {self.ncols} columns")

        for i, col in enumerate(self.columns):
            col.append(row[i])

    def rows(self) -> List[List[Field]]:
        """
        Returns a list of rows in the table. Each row is a list of fields.
        """
        return [[col[i] for col in self.columns] for i in range(len(self))]

    def __len__(self) -> int:
        return self.nrows

    def format_lines(self, width: Optional[int] = None) -> Iterator[str]:
        """
        Iterator over formatted lines of the table. Fields may span multiple lines if
        they are set to wrap instead of truncate.

        :param width: Width to fit the table.
        :returns: Iterator over lines which can be printed to the terminal.
        """

        # get terminal width if no width is given
        if not width:
            width, height = click.get_terminal_size()

        available_width = width - self.padding * len(self.columns)
        raw_col_widths = [col.display_width for col in self.columns]

        # Allocate column width from available width,
        # weighted by the raw width of each column.
        n = 3
        sum_widths = sum(w ** n for w in raw_col_widths)
        subtract = max([sum(raw_col_widths) - available_width, 0])
        allocated_col_widths = tuple(
            round(w - subtract * w ** n / sum_widths) for w in raw_col_widths
        )

        spacer = " " * self.padding

        # generate line for titles
        if any(col.has_title for col in self.columns):
            titles: List[str] = []

            for col, width in zip(self.columns, allocated_col_widths):
                if col.title:
                    titles.append(col.title.format(width)[0])
                else:
                    titles.append(adjust("", width))

            line = spacer.join(titles)
            yield line.rstrip()

        # generate lines for rows
        for row in self.rows():
            cells = []

            for field, alloc_width in zip(row, allocated_col_widths):
                cells.append(field.format(alloc_width))

            n_lines = max(len(cell) for cell in cells)

            for i in range(n_lines):

                line_parts = []

                for cell_lines, alloc_width in zip(cells, allocated_col_widths):
                    try:
                        line_parts.append(cell_lines[i])
                    except IndexError:
                        line_parts.append(adjust("", width=alloc_width))

                line = spacer.join(line_parts)
                yield line.rstrip()

    def format(self, width: Optional[int] = None) -> str:
        """
        Returns a fully formatted table as a string with linebreaks.

        :param width: Width to fit the table.
        :returns: Formatted table.
        """
        return "\n".join(self.format_lines(width))

    def echo(self):
        """Prints the table to the terminal."""
        for line in self.format_lines():
            click.echo(line)


class Grid:
    """
    A grid of fields which can be printed to stdout.

    :param fields: A sequence of fields (strings, datetimes, any objects with a string
        representation).
    :param padding: Padding between fields.
    :param align: Alignment of strings in the grid.
    """

    fields: List[Field]

    def __init__(
        self, fields: Sequence = (), padding: int = 2, align: Align = Align.Left
    ):

        self.fields = []
        self.padding = padding
        self.align = align

        for field in fields:
            self.fields.append(self._to_field(field))

    def append(self, field: Any) -> None:
        """Appends a field to the grid."""
        self.fields.append(self._to_field(field))

    def __iter__(self) -> Iterator:
        return iter(self.fields)

    def __len__(self) -> int:
        return len(self.fields)

    def _to_field(self, field: Any) -> Field:

        from datetime import datetime

        if isinstance(field, Field):
            return field
        elif isinstance(field, datetime):
            return DateField(field)
        else:
            return TextField(str(field), align=self.align)

    def format_lines(self, width: Optional[int] = None) -> Iterator[str]:
        """
        Iterator over formatted lines of the grid.

        :param width: Width to fit the grid.
        :returns: Iterator over lines which can be printed to the terminal.
        """

        if len(self.fields) > 0:

            from . import chunks

            # get terminal width if no width is given
            if not width:
                width, height = click.get_terminal_size()

            field_width = max(field.display_width for field in self.fields)
            field_width = min(field_width, width)  # cap at terminal / total width
            field_texts = [field.format(field_width)[0] for field in self.fields]

            n_columns = max(width // (field_width + self.padding), 1)

            rows = chunks(field_texts, n_columns)
            spacer = " " * self.padding

            for row in rows:
                line = spacer.join(row)
                yield line.rstrip()

        else:
            yield ""

    def format(self, width: Optional[int] = None) -> str:
        """
        Returns a fully formatted grid as a string with linebreaks.

        :param width: Width to fit the table.
        :returns: Formatted grid.
        """
        return "\n".join(self.format_lines(width))

    def echo(self):
        """Prints the grid to the terminal."""
        for line in self.format_lines():
            click.echo(line)


# ==== interactive prompts =============================================================


def echo(message: str, nl: bool = True, prefix: Prefix = Prefix.NONE) -> None:

    if prefix is Prefix.Ok:
        pre = click.style("✓", fg="green") + " "
    elif prefix is Prefix.Warn:
        pre = click.style("!", fg="red") + " "
    elif prefix is Prefix.Info:
        pre = "- "
    else:
        pre = ""

    click.echo(f"{pre}{message}", nl=nl)


def info(message: str, nl: bool = True) -> None:
    echo(message, nl=nl, prefix=Prefix.Info)


def warn(message: str, nl: bool = True) -> None:
    echo(message, nl=nl, prefix=Prefix.Warn)


def ok(message: str, nl: bool = True) -> None:
    echo(message, nl=nl, prefix=Prefix.Ok)


def _style_message(message: str) -> str:
    pre = click.style("?", fg="green")
    return f"{pre} {message} "


def _syle_hint(hint: str) -> str:
    return click.style(hint, fg="white") + " " if hint else ""


orange = "\x1b[38;5;214m"
cyan = "\x1b[38;5;6m"
grey = "\x1b[90m"
bold = "\x1b[1m"

response_color = cyan
focus_color = f"{response_color}"


class loading:

    _animation = ("...", "   ", ".  ", ".. ")

    def __init__(self, iterable, prefix="Loading", animation=None, clear=True):

        import itertools

        self.iterable = iterable
        self.prefix = prefix
        self.clear = clear
        self.indicator = itertools.cycle(animation or loading._animation)

    def _render(self) -> None:
        click.echo(self.prefix + next(self.indicator) + "\r", nl=False)

    def __enter__(self) -> "loading":
        self._render()
        return self

    def __iter__(self) -> "loading":
        return self

    def __next__(self) -> Any:
        res = next(self.iterable)
        self._render()
        return res

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if self.clear:
            click.echo(" " * (len(self.prefix) + 3) + "\r", nl=False)
        else:
            click.echo("")


def prompt(message: str, default: str = "", validate: Optional[Callable] = None) -> str:

    import survey

    styled_default = _syle_hint(default)
    styled_message = _style_message(message)

    def view(value: str) -> Tuple[str]:
        response = value or default
        return (response,)

    def check(value: str) -> bool:
        if validate is None:
            return True
        elif value == "" and default:
            return True
        else:
            return validate(value)

    res = survey.input(
        styled_message,
        hint=styled_default,
        view=view,
        check=check,
        color=response_color,
    )

    return res


def confirm(message: str, default: Optional[bool] = True) -> bool:

    import survey

    styled_message = _style_message(message)

    return survey.confirm(styled_message, default=default, color=response_color)


def select(message: str, options: Sequence[str], hint="") -> int:

    import survey

    try:
        styled_hint = _syle_hint(hint)
        styled_message = _style_message(message)

        index = survey.select(
            options,
            styled_message,
            focus=focus_color,
            color=response_color,
            hint=styled_hint,
        )

        return index
    except (KeyboardInterrupt, SystemExit):
        survey.api.respond()
        raise


def select_multiple(message: str, options: Sequence[str], hint="") -> List[int]:

    import survey

    try:
        styled_hint = _syle_hint(hint)
        styled_message = _style_message(message)

        kwargs = {"hint": styled_hint} if hint else {}

        def view(value: Sequence[int]) -> Tuple[str]:

            chosen = [options[index] for index in value]
            response = ", ".join(chosen)

            if len(value) == 0 or len(response) > 50:
                response = f"[{len(value)} chosen]"

            return (response,)

        indices = survey.select(
            options,
            styled_message,
            multi=True,
            focus=focus_color,
            color=response_color,
            pin="[✓] ",
            unpin="[ ] ",
            view=view,
            **kwargs,
        )

        return indices

    except (KeyboardInterrupt, SystemExit):
        survey.api.respond()
        raise


def select_path(
    message: str,
    default: str = "",
    validate: Callable = lambda x: True,
    exists: bool = False,
    only_directories: bool = False,
) -> str:

    import os

    import survey

    styled_default = _syle_hint(f"[{default}]")
    styled_message = _style_message(message)

    def view(value: str) -> Tuple[str]:
        response = value or default
        return (response,)

    failed = False

    def check(value: str) -> bool:

        nonlocal failed

        if value == "" and default:
            return True

        full_path = os.path.expanduser(value)
        dir_condition = os.path.isdir(full_path) or not only_directories
        exist_condition = os.path.exists(full_path) or not exists

        if not dir_condition:
            survey.update(click.style("(not a directory) ", fg="red"))
        elif not exist_condition:
            survey.update(click.style("(does not exist) ", fg="red"))

        passed = dir_condition and exist_condition and validate(value)
        failed = not passed

        return passed

    def callback(event: str, result: str, *args) -> None:
        nonlocal failed

        if event == "delete" and failed:
            survey.update(styled_default)
            failed = False

    res = survey.input(
        styled_message,
        hint=styled_default,
        view=view,
        check=check,
        callback=callback,
        color=response_color,
    )

    return res or default


class RemoteApiError(click.ClickException):
    def __init__(self, title: str, message: str) -> None:
        super().__init__(message)
        self.title = title

    def format_message(self) -> str:
        return f"{self.title}. {self.message}"

    def show(self, file=None) -> None:
        warn(self.format_message())


class CliException(click.ClickException):
    def show(self, file=None) -> None:
        warn(self.format_message())



# -*- coding: utf-8 -*-
"""This module defines custom logging records and handlers."""

import logging
from collections import deque
from concurrent.futures._base import Future, wait
from typing import Deque, Optional, List

try:
    from concurrent.futures import InvalidStateError  # type: ignore
except ImportError:
    # Python 3.7 and lower
    InvalidStateError = RuntimeError  # type: ignore

import sdnotify  # type: ignore

try:
    from systemd import journal  # type: ignore
except ImportError:
    journal = None

from .utils import sanitize_string


if journal:

    def safe_journal_sender(MESSAGE: str, **kwargs) -> None:

        MESSAGE = sanitize_string(MESSAGE)

        for key, value in kwargs.items():
            if isinstance(value, str):
                kwargs[key] = sanitize_string(value)

        journal.send(MESSAGE, **kwargs)


else:

    def safe_journal_sender(MESSAGE: str, **kwargs) -> None:
        pass


class EncodingSafeLogRecord(logging.LogRecord):
    """A log record which ensures that messages contain only unicode characters

    This is useful when log messages may contain file paths generates by OS APIs. In
    Python, such path strings may contain surrogate escapes and will therefore raise
    :class:`UnicodeEncodingError`s under many circumstances (printing to stdout, etc).
    """

    def getMessage(self) -> str:
        """
        Formats the log message and replaces all surrogate escapes with "�".
        """
        msg = super().getMessage()
        return sanitize_string(msg)


logging.setLogRecordFactory(EncodingSafeLogRecord)


class CachedHandler(logging.Handler):
    """Handler which stores past records

    This is used to populate Maestral's status and error interfaces. The method
    :meth:`wait_for_emit` can be used from another thread to block until a new record is
    emitted, for instance to react to state changes.

    :param level: Initial log level. Defaults to NOTSET.
    :param maxlen: Maximum number of records to store. If ``None``, all records will be
        stored. Defaults to ``None``.
    """

    cached_records: Deque[logging.LogRecord]
    _emit_future: Future

    def __init__(
        self, level: int = logging.NOTSET, maxlen: Optional[int] = None
    ) -> None:
        super().__init__(level=level)
        self.cached_records = deque([], maxlen)
        self._emit_future = Future()

    def emit(self, record: logging.LogRecord) -> None:
        """
        Logs the specified log record and saves it to the cache.

        :param record: Log record.
        """
        self.cached_records.append(record)

        # notify any waiting coroutines that we have a status change
        try:
            self._emit_future.set_result(True)
        except InvalidStateError:
            pass

    def wait_for_emit(self, timeout: Optional[float]) -> bool:
        """
        Blocks until a new record is emitted.

        :param timeout: Maximum time to block before returning.
        :returns: ``True``if there was a status change, ``False`` in case of a timeout.
        """
        done, not_done = wait([self._emit_future], timeout=timeout)
        self._emit_future = Future()  # reset future
        return len(done) == 1

    def getLastMessage(self) -> str:
        """
        :returns: The log message of the last record or an empty string.
        """
        try:
            last_record = self.cached_records[-1]
            return last_record.getMessage()
        except IndexError:
            return ""

    def getAllMessages(self) -> List[str]:
        """
        :returns: A list of all record messages.
        """
        return [r.getMessage() for r in self.cached_records]

    def clear(self) -> None:
        """
        Clears all cached records.
        """
        self.cached_records.clear()


class SdNotificationHandler(logging.Handler):
    """Handler which emits messages as systemd notifications

    This is useful when used from a systemd service and will do nothing when no
    NOTIFY_SOCKET is provided.

    :param level: Initial log level. Defaults to NOTSET.
    """

    notifier = sdnotify.SystemdNotifier()

    def emit(self, record: logging.LogRecord) -> None:
        """
        Sends the record massage to systemd as service status.

        :param record: Log record.
        """
        self.notifier.notify(f"STATUS={record.message}")



# -*- coding: utf-8 -*-
"""
Notification backend for Linux. Includes an implementation to send desktop notifications
over Dbus. Responding to user interaction with a notification requires a running asyncio
event loop.
"""

# system imports
import asyncio
import logging
from typing import Optional, Type, Coroutine

# external imports
from dbus_next import Variant  # type: ignore
from dbus_next.aio import MessageBus, ProxyInterface  # type: ignore

# local imports
from .notify_base import Notification, DesktopNotifierBase, NotificationLevel


__all__ = ["Impl", "DBusDesktopNotifier"]

logger = logging.getLogger(__name__)

Impl: Optional[Type[DesktopNotifierBase]]


class DBusDesktopNotifier(DesktopNotifierBase):
    """DBus notification backend for Linux

    This implements the org.freedesktop.Notifications standard. The DBUS connection is
    created in a thread with a running asyncio loop to handle clicked notifications.
    """

    _to_native_urgency = {
        NotificationLevel.Low: Variant("y", 0),
        NotificationLevel.Normal: Variant("y", 1),
        NotificationLevel.Critical: Variant("y", 2),
    }

    def __init__(self, app_name: str, app_id: str) -> None:
        super().__init__(app_name, app_id)
        self._loop = asyncio.get_event_loop()
        self.interface: Optional[ProxyInterface] = None
        self._force_run_in_loop(self._init_dbus())

    def _force_run_in_loop(self, coro: Coroutine) -> None:

        if self._loop.is_running():
            asyncio.run_coroutine_threadsafe(coro, self._loop)
        else:
            self._loop.run_until_complete(coro)

    async def _init_dbus(self) -> None:

        try:
            self.bus = await MessageBus().connect()
            introspection = await self.bus.introspect(
                "org.freedesktop.Notifications", "/org/freedesktop/Notifications"
            )
            self.proxy_object = self.bus.get_proxy_object(
                "org.freedesktop.Notifications",
                "/org/freedesktop/Notifications",
                introspection,
            )
            self.interface = self.proxy_object.get_interface(
                "org.freedesktop.Notifications"
            )
            self.interface.on_action_invoked(self._on_action)
        except Exception:
            self.interface = None
            logger.warning("Could not connect to DBUS interface", exc_info=True)

    def send(self, notification: Notification) -> None:
        """
        Sends a notification.

        :param notification: Notification to send.
        """
        self._force_run_in_loop(self._send(notification))

    async def _send(self, notification: Notification) -> None:

        # Do nothing if we couldn't connect.
        if not self.interface:
            return

        # Get an internal ID for the notifications. This will recycle an old ID if we
        # are above the max number of notifications.
        internal_nid = self._next_nid()

        # Get the old notification to replace, if any.
        notification_to_replace = self.current_notifications.get(internal_nid)

        if notification_to_replace:
            replaces_nid = notification_to_replace.identifier
        else:
            replaces_nid = 0

        # Create list of actions with default and user-supplied.
        actions = ["default", "default"]

        for button_name in notification.buttons.keys():
            actions += [button_name, button_name]

        try:
            # Post the new notification and record the platform ID assigned to it.
            platform_nid = await self.interface.call_notify(
                self.app_name,  # app_name
                replaces_nid,  # replaces_id
                notification.icon or "",  # app_icon
                notification.title,  # summary
                notification.message,  # body
                actions,  # actions
                {"urgency": self._to_native_urgency[notification.urgency]},  # hints
                -1,  # expire_timeout (-1 = default)
            )
        except Exception:
            # This may fail for several reasons: there may not be a systemd service
            # file for 'org.freedesktop.Notifications' or the system configuration
            # may have changed after DesktopNotifierFreedesktopDBus was initialized.
            logger.warning("Notification failed", exc_info=True)
        else:
            # Store the notification for future replacement and to keep track of
            # user-supplied callbacks.
            notification.identifier = platform_nid
            self.current_notifications[internal_nid] = notification

    def _on_action(self, nid, action_key) -> None:

        # Get the notification instance from the platform ID.
        nid = int(nid)
        action_key = str(action_key)
        notification = next(
            iter(n for n in self.current_notifications.values() if n.identifier == nid),
            None,
        )

        # Execute any callbacks for button clicks.
        if notification:
            if action_key == "default" and notification.action:
                notification.action()
            else:
                callback = notification.buttons.get(action_key)

                if callback:
                    callback()


Impl = DBusDesktopNotifier



# -*- coding: utf-8 -*-
"""
This module contains functions to retrieve platform dependent locations to store app
data. It supports macOS and Linux.
"""

# system imports
import os
import os.path as osp
import platform
from typing import Optional

# local imports
from ..config.base import (
    get_home_dir,
    get_conf_path,
    get_data_path,
    to_full_path,
)


__all__ = [
    "get_home_dir",
    "get_conf_path",
    "get_data_path",
    "get_log_path",
    "get_cache_path",
    "get_autostart_path",
    "get_runtime_path",
]

home_dir = get_home_dir()


def get_cache_path(
    subfolder: Optional[str] = None, filename: Optional[str] = None, create: bool = True
) -> str:
    """
    Returns the default cache path for the platform. This will be:

        - macOS: "~/Library/Application Support/SUBFOLDER/FILENAME"
        - Linux: "$XDG_CACHE_HOME/SUBFOLDER/FILENAME"
        - fallback: "$HOME/.cache/SUBFOLDER/FILENAME"

    :param subfolder: The subfolder for the app.
    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """
    if platform.system() == "Darwin":
        cache_path = get_conf_path(create=False)
    elif platform.system() == "Linux":
        fallback = osp.join(home_dir, ".cache")
        cache_path = os.environ.get("XDG_CACHE_HOME", fallback)
    else:
        raise RuntimeError("Platform not supported")

    return to_full_path(cache_path, subfolder, filename, create)


def get_log_path(
    subfolder: Optional[str] = None, filename: Optional[str] = None, create: bool = True
) -> str:
    """
    Returns the default log path for the platform. This will be:

        - macOS: "~/Library/Logs/SUBFOLDER/FILENAME"
        - Linux: "$XDG_CACHE_HOME/SUBFOLDER/FILENAME"
        - fallback: "$HOME/.cache/SUBFOLDER/FILENAME"

    :param subfolder: The subfolder for the app.
    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """

    if platform.system() == "Darwin":
        log_path = osp.join(home_dir, "Library", "Logs")
    elif platform.system() == "Linux":
        log_path = get_cache_path(create=False)
    else:
        raise RuntimeError("Platform not supported")

    return to_full_path(log_path, subfolder, filename, create)


def get_autostart_path(filename: Optional[str] = None, create: bool = True) -> str:
    """
    Returns the default path for login items for the platform. This will be:

        - macOS: "~/Library/LaunchAgents/FILENAME"
        - Linux: "$XDG_CONFIG_HOME/autostart/FILENAME"
        - fallback: "$HOME/.config/autostart/FILENAME"

    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """
    if platform.system() == "Darwin":
        autostart_path = osp.join(home_dir, "Library", "LaunchAgents")
    elif platform.system() == "Linux":
        autostart_path = get_conf_path("autostart", create=create)
    else:
        raise RuntimeError("Platform not supported")

    if filename:
        autostart_path = osp.join(autostart_path, filename)

    return autostart_path


def get_runtime_path(
    subfolder: Optional[str] = None, filename: Optional[str] = None, create: bool = True
) -> str:
    """
    Returns the default runtime path for the platform. This will be:

        - macOS: "~/Library/Application Support/SUBFOLDER/FILENAME"
        - Linux: "$XDG_RUNTIME_DIR/SUBFOLDER/FILENAME"
        - fallback: "$HOME/.cache/SUBFOLDER/FILENAME"

    :param subfolder: The subfolder for the app.
    :param filename: The filename to append for the app.
    :param create: If ``True``, the folder ``subfolder`` will be created on-demand.
    """

    if platform.system() == "Darwin":
        runtime_path = get_conf_path(create=False)
    elif platform.system() == "Linux":
        fallback = get_cache_path(create=False)
        runtime_path = os.environ.get("XDG_RUNTIME_DIR", fallback)
    else:
        raise RuntimeError("Platform not supported")

    return to_full_path(runtime_path, subfolder, filename, create)



# -*- coding: utf-8 -*-

# system imports
from setuptools import setup, find_packages  # type: ignore


# proceed with actual install
install_requires = [
    "alembic>=1.3,<1.5",
    "bugsnag>=3.4,<5.0",
    "click>=7.1.1,<8.0",
    "dropbox>=10.9.0,<12.0",
    "dbus-next>=0.1.4;sys_platform=='linux'",
    "fasteners>=0.15",
    "importlib_metadata;python_version<'3.8'",
    "importlib_resources;python_version<'3.9'",
    "keyring>=19,<22",
    "keyrings.alt>=3.1.0,<5.0",
    "packaging",
    "pathspec>=0.5.8",
    "Pyro5>=5.10",
    "requests>=2.16.2",
    "rubicon-objc>=0.3.1;sys_platform=='darwin'",
    "sdnotify>=0.3.2",
    "setuptools",
    "sqlalchemy>=1.3,<1.4",
    "survey>=2.1.0,<3.0",
    "watchdog>=0.10.0,<=0.10.3",
]

gui_requires = [
    "maestral-qt>=1.3.0;sys_platform=='linux'",
    "maestral-cocoa>=1.3.0;sys_platform=='darwin'",
]

syslog_requires = ["systemd-python"]

dev_requires = [
    "black",
    "bump2version",
    "flake8",
    "mypy",
    "pre-commit",
    "pytest",
    "pytest-cov",
]

docs_require = [
    "sphinx",
    "m2r2",
    "sphinx-autoapi",
    "sphinx_rtd_theme",
]

setup(
    name="maestral",
    author="Sam Schott",
    author_email="ss2151@cam.ac.uk",
    version="1.3.0",
    url="https://github.com/SamSchott/maestral",
    description="Open-source Dropbox client for macOS and Linux.",
    license="MIT",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={
        "maestral": ["resources/*"],
    },
    setup_requires=["wheel"],
    install_requires=install_requires,
    extras_require={
        "gui": gui_requires,
        "syslog": syslog_requires,
        "dev": dev_requires,
        "docs": docs_require,
    },
    zip_safe=False,
    entry_points={
        "console_scripts": ["maestral=maestral.cli:main"],
        "pyinstaller40": ["hook-dirs=maestral.__pyinstaller:get_hook_dirs"],
    },
    python_requires=">=3.6",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    data_files=[
        ("share/icons/hicolor/512x512/apps", ["src/maestral/resources/maestral.png"])
    ],
)



# -*- coding: utf-8 -*-
"""Module for content hashing."""

# system imports
import hashlib


class DropboxContentHasher:
    """
    Computes a hash using the same algorithm that the Dropbox API uses for the
    the "content_hash" metadata field.
    The digest() method returns a raw binary representation of the hash.  The
    hexdigest() convenience method returns a hexadecimal-encoded version, which
    is what the "content_hash" metadata field uses.
    This class has the same interface as the hashers in the standard 'hashlib'
    package.

    Example:
        hasher = DropboxContentHasher()
        with open('some-file', 'rb') as f:
            while True:
                chunk = f.read(1024)  # or whatever chunk size you want
                if len(chunk) == 0:
                    break
                hasher.update(chunk)
        print(hasher.hexdigest())
    """

    BLOCK_SIZE = 4 * 1024 * 1024

    def __init__(self):
        self._overall_hasher = hashlib.sha256()
        self._block_hasher = hashlib.sha256()
        self._block_pos = 0

        self.digest_size = self._overall_hasher.digest_size

    def update(self, new_data):
        if self._overall_hasher is None:
            raise RuntimeError(
                "can't use this object anymore; you already called digest()"
            )

        if not isinstance(new_data, bytes):
            raise ValueError("Expecting a byte string, got {!r}".format(new_data))

        new_data_pos = 0
        while new_data_pos < len(new_data):
            if self._block_pos == self.BLOCK_SIZE:
                self._overall_hasher.update(self._block_hasher.digest())
                self._block_hasher = hashlib.sha256()
                self._block_pos = 0

            space_in_block = self.BLOCK_SIZE - self._block_pos
            part = new_data[new_data_pos : (new_data_pos + space_in_block)]
            self._block_hasher.update(part)

            self._block_pos += len(part)
            new_data_pos += len(part)

    def _finish(self):
        if self._overall_hasher is None:
            raise RuntimeError(
                "can't use this object anymore; you already called digest() or hexdigest()"
            )

        if self._block_pos > 0:
            self._overall_hasher.update(self._block_hasher.digest())
            self._block_hasher = None
        h = self._overall_hasher
        self._overall_hasher = None  # Make sure we can't use this object anymore.
        return h

    def digest(self):
        return self._finish().digest()

    def hexdigest(self):
        return self._finish().hexdigest()

    def copy(self):
        c = DropboxContentHasher.__new__(DropboxContentHasher)
        c._overall_hasher = self._overall_hasher.copy()
        c._block_hasher = self._block_hasher.copy()
        c._block_pos = self._block_pos
        return c


class StreamHasher:
    """
    A wrapper around a file-like object (either for reading or writing)
    that hashes everything that passes through it.  Can be used with
    DropboxContentHasher or any 'hashlib' hasher.
    Example:
        hasher = DropboxContentHasher()
        with open('some-file', 'rb') as f:
            wrapped_f = StreamHasher(f, hasher)
            response = some_api_client.upload(wrapped_f)
        locally_computed = hasher.hexdigest()
        assert response.content_hash == locally_computed
    """

    def __init__(self, f, hasher):
        self._f = f
        self._hasher = hasher

    def close(self):
        return self._f.close()

    def flush(self):
        return self._f.flush()

    def fileno(self):
        return self._f.fileno()

    def tell(self):
        return self._f.tell()

    def read(self, *args):
        b = self._f.read(*args)
        self._hasher.update(b)
        return b

    def write(self, b):
        self._hasher.update(b)
        return self._f.write(b)

    def next(self):
        b = self._f.next()
        self._hasher.update(b)
        return b

    def readline(self, *args):
        b = self._f.readline(*args)
        self._hasher.update(b)
        return b

    def readlines(self, *args):
        bs = self._f.readlines(*args)
        for b in bs:
            self._hasher.update(b)
        return b



# -*- coding: utf-8 -*-
"""
This file defines the functions to configure and interact with Maestral from the command
line. Some imports are deferred to the functions that required them in order to reduce
the startup time of individual CLI commands.
"""

# system imports
import sys
import os
import os.path as osp
import functools
import time
from typing import Optional, Dict, List, Tuple, Callable, Union, cast, TYPE_CHECKING

# external imports
import click
import Pyro5.errors  # type: ignore

# local imports
from . import __version__, __author__, __url__
from .utils import cli

if TYPE_CHECKING:
    from .main import Maestral
    from .daemon import MaestralProxy


# ======================================================================================
# CLI dialogs and helper functions
# ======================================================================================

OK = click.style("[OK]", fg="green")
FAILED = click.style("[FAILED]", fg="red")
KILLED = click.style("[KILLED]", fg="red")


def stop_daemon_with_cli_feedback(config_name: str) -> None:
    """Wrapper around :meth:`daemon.stop_maestral_daemon_process`
    with command line feedback."""

    from .daemon import stop_maestral_daemon_process, Stop

    click.echo("Stopping Maestral...", nl=False)
    res = stop_maestral_daemon_process(config_name)
    if res == Stop.Ok:
        click.echo(" " * 8 + OK)
    elif res == Stop.NotRunning:
        click.echo("\rMaestral daemon is not running.")
    elif res == Stop.Killed:
        click.echo(" " * 8 + KILLED)
    elif res == Stop.Failed:
        click.echo(" " * 8 + FAILED)


def select_dbx_path_dialog(
    config_name: str, default_dir_name: Optional[str] = None, allow_merge: bool = False
) -> str:
    """
    A CLI dialog to ask for a local Dropbox folder location.

    :param config_name: The configuration to use for the default folder name.
    :param default_dir_name: The default directory name. Defaults to
        "Dropbox ({config_name})" if not given.
    :param allow_merge: If ``True``, allows the selection of an existing folder without
        deleting it. Defaults to ``False``.
    :returns: Path given by user.
    """

    from .utils.path import delete

    default_dir_name = default_dir_name or f"Dropbox ({config_name.capitalize()})"

    while True:
        res = cli.select_path(
            "Please choose a local Dropbox folder:",
            default=f"~/{default_dir_name}",
            only_directories=True,
        )
        res = res.rstrip(osp.sep)

        dropbox_path = osp.expanduser(res)

        if osp.exists(dropbox_path):
            if allow_merge:
                text = (
                    "Directory already exists. Do you want to replace it "
                    "or merge its content with your Dropbox?"
                )
                choice = cli.select(text, options=["replace", "merge", "cancel"])
            else:
                text = (
                    "Directory already exists. Do you want to replace it? "
                    "Its content will be lost!"
                )
                replace = cli.confirm(text)
                choice = 0 if replace else 2

            if choice == 0:
                err = delete(dropbox_path)
                if err:
                    cli.warn(
                        "Could not write to selected location. "
                        "Please make sure that you have sufficient permissions."
                    )
                else:
                    cli.ok("Replaced existing folder")
                    return dropbox_path
            elif choice == 1:
                cli.ok("Merging with existing folder")
                return dropbox_path

        else:
            return dropbox_path


def link_dialog(m: Union["MaestralProxy", "Maestral"]) -> None:
    """
    A CLI dialog for linking a Dropbox account.

    :param m: Proxy to Maestral daemon.
    """

    authorize_url = m.get_auth_url()

    cli.info(f"Linking new account for '{m.config_name}' config")
    cli.info("Retrieving auth code from Dropbox")
    choice = cli.select(
        "How would you like to you link your account?",
        options=["Open Dropbox website", "Print auth URL to console"],
    )

    if choice == 0:
        click.launch(authorize_url)
    else:
        cli.info("Open the URL below to retrieve an auth code:")
        cli.info(authorize_url)

    res = -1
    while res != 0:
        auth_code = cli.prompt("Enter the auth code:")
        auth_code = auth_code.strip()

        res = m.link(auth_code)

        if res == 0:
            email = m.get_state("account", "email")
            cli.ok(f"Linked to {email}")
        elif res == 1:
            cli.warn("Invalid token, please try again")
        elif res == 2:
            cli.warn(" Could not connect to Dropbox, please try again")


def check_for_updates() -> None:
    """
    Checks if updates are available by reading the cached release number from the
    config file and notifies the user. Prints an update note to the command line.
    """
    from packaging.version import Version
    from .config import MaestralConfig, MaestralState

    conf = MaestralConfig("maestral")
    state = MaestralState("maestral")

    interval = conf.get("app", "update_notification_interval")
    last_update_check = state.get("app", "update_notification_last")
    latest_release = state.get("app", "latest_release")

    if interval == 0 or time.time() - last_update_check < interval:
        return

    has_update = Version(__version__) < Version(latest_release)

    if has_update:
        cli.echo(
            f"Update available v{__version__} → v{latest_release}. "
            f"Please use your package manager to update."
        )


def check_for_fatal_errors(m: Union["MaestralProxy", "Maestral"]) -> bool:
    """
    Checks the given Maestral instance for fatal errors such as revoked Dropbox access,
    deleted Dropbox folder etc. Prints a nice representation to the command line.

    :param m: Proxy to Maestral daemon or Maestral instance.
    :returns: True in case of fatal errors, False otherwise.
    """

    import textwrap

    maestral_err_list = m.fatal_errors

    if len(maestral_err_list) > 0:

        width, height = click.get_terminal_size()

        err = maestral_err_list[0]
        err_title = cast(str, err["title"])
        err_msg = cast(str, err["message"])

        wrapped_msg = textwrap.fill(err_msg, width=width)

        click.echo("")
        click.secho(err_title, fg="red")
        click.secho(wrapped_msg, fg="red")
        click.echo("")

        return True
    else:
        return False


def catch_maestral_errors(func: Callable) -> Callable:
    """
    Decorator that catches a MaestralApiError and prints it as a useful message to the
    command line instead of printing the full stacktrace.
    """

    from .errors import MaestralApiError

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except MaestralApiError as exc:
            raise cli.RemoteApiError(exc.title, exc.message)
        except ConnectionError:
            raise cli.CliException("Could not connect to Dropbox")

    return wrapper


# ======================================================================================
# Custom parameter types
# ======================================================================================

# A custom parameter:
# * needs a name
# * needs to pass through None unchanged
# * needs to convert from a string
# * needs to convert its result type through unchanged (eg: needs to be idempotent)
# * needs to be able to deal with param and context being None. This can be the case
#   when the object is used with prompt inputs.


class DropboxPath(click.ParamType):
    """A command line parameter representing a Dropbox path

    :param file_okay: Controls if a file is a possible value.
    :param dir_okay: Controls if a directory is a possible value.
    """

    name = "Dropbox path"
    envvar_list_splitter = osp.pathsep

    def __init__(self, file_okay: bool = True, dir_okay: bool = True) -> None:
        self.file_okay = file_okay
        self.dir_okay = dir_okay

    #
    # def shell_complete(
    #     self,
    #     ctx: Optional[click.Context],
    #     param: Optional[click.Parameter],
    #     incomplete: str,
    # ) -> List["CompletionItem"]:
    #
    #     from click.shell_completion import CompletionItem
    #     from .utils import removeprefix
    #
    #     matches: List[str] = []
    #
    #     # check if we have been given an absolute path
    #     incomplete = incomplete.lstrip("/")
    #
    #     # get the Maestral config for which to complete paths
    #     try:
    #         config_name = ctx.params["config_name"]
    #     except (KeyError, AttributeError):
    #         # attribute error occurs when ctx = None
    #         config_name = "maestral"
    #
    #     # get all matching paths in our local Dropbox folder
    #     # TODO: query from server if not too slow
    #
    #     config = MaestralConfig(config_name)
    #     dropbox_dir = config.get("main", "path")
    #     local_incomplete = osp.join(dropbox_dir, incomplete)
    #     local_dirname = osp.dirname(local_incomplete)
    #
    #     if osp.isdir(local_dirname):
    #
    #         with os.scandir(local_dirname) as it:
    #             for entry in it:
    #                 if entry.path.startswith(local_incomplete):
    #                     if entry.is_dir() and self.dir_okay:
    #                         dbx_path = removeprefix(entry.path, dropbox_dir)
    #                         matches.append(dbx_path + "/")
    #                     elif entry.is_file() and self.file_okay:
    #                         dbx_path = removeprefix(entry.path, dropbox_dir)
    #                         matches.append(dbx_path)
    #
    #     # get all matching excluded items
    #
    #     for dbx_path in config.get("main", "excluded_items"):
    #         if dbx_path.startswith("/" + incomplete):
    #             matches.append(dbx_path)
    #
    #     return [CompletionItem(m.lstrip("/")) for m in matches]


class ConfigName(click.ParamType):
    """ "A command line parameter representing a Dropbox path

    :param existing: If ``True`` require an existing config, otherwise create a new
        config on demand.
    """

    name = "config"

    def __init__(self, existing: bool = True) -> None:
        self.existing = existing

    def convert(
        self,
        value: Optional[str],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> Optional[str]:

        if value is None:
            return value

        from .config import validate_config_name, list_configs

        if not self.existing:

            # accept all valid config names
            try:
                return validate_config_name(value)
            except ValueError:
                raise cli.CliException(
                    "Configuration name may not contain any whitespace"
                )

        else:

            # accept only existing config names
            if value in list_configs():
                return value
            else:
                raise cli.CliException(
                    f"Configuration '{value}' does not exist. "
                    f"Use 'maestral configs' to list all configurations."
                )

    #
    # def shell_complete(
    #     self,
    #     ctx: Optional[click.Context],
    #     param: Optional[click.Parameter],
    #     incomplete: str,
    # ) -> List["CompletionItem"]:
    #
    #     matches = [conf for conf in list_configs() if conf.startswith(incomplete)]
    #     return [CompletionItem(m) for m in matches]
    #


# ======================================================================================
# Command groups
# ======================================================================================


class OrderedGroup(click.Group):
    """Click command group with customizable order of help output."""

    def command(self, *args, **kwargs) -> Callable:
        """Behaves the same as :meth:`click.Group.command()` except captures a section
        name for listing command names in help.
        """
        section = kwargs.pop("section", "Commands")

        from click.decorators import command

        def decorator(f):
            cmd = command(*args, **kwargs)(f)
            cmd.section = section
            self.add_command(cmd)

            return cmd

        return decorator

    def group(self, *args, **kwargs) -> Callable:
        """Behaves the same as :meth:`click.Group.group()` except captures a section
        name for listing command names in help.
        """
        section = kwargs.pop("section", "Commands")

        from click.decorators import group

        def decorator(f):
            cmd = group(*args, **kwargs)(f)
            cmd.section = section
            self.add_command(cmd)

            return cmd

        return decorator

    def format_commands(
        self, ctx: click.Context, formatter: click.HelpFormatter
    ) -> None:
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        commands = []

        for name in self.commands:
            cmd = self.get_command(ctx, name)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((name, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            max_len = max(len(name) for name, cmd in commands)
            limit = formatter.width - 6 - max_len  # type: ignore

            sections: Dict[str, List[Tuple[str, click.Command]]] = {}

            # group commands into sections
            for name, cmd in commands:
                try:
                    sections[cmd.section].append((name, cmd))  # type: ignore
                except KeyError:
                    sections[cmd.section] = [(name, cmd)]  # type: ignore

            # format sections individually
            for section, cmds in sections.items():

                rows = []

                for name, cmd in cmds:
                    name = name.ljust(max_len)
                    help = cmd.get_short_help_str(limit)
                    rows.append((name, help))

                if rows:
                    with formatter.section(section):
                        formatter.write_dl(rows)


@click.group(
    cls=OrderedGroup,
    context_settings={"help_option_names": ["-h", "--help"]},
    invoke_without_command=True,
    no_args_is_help=True,
    help="Dropbox client for Linux and macOS.",
)
@click.option(
    "--version",
    "-V",
    is_flag=True,
    default=False,
    help="Show version and exit.",
)
def main(version: bool):

    if version:
        click.echo(__version__)


# ======================================================================================
# Core commands
# ======================================================================================

config_option = click.option(
    "-c",
    "--config-name",
    default="maestral",
    type=ConfigName(existing=False),
    is_eager=True,
    expose_value=True,
    help="Run command with the given configuration.",
)

existing_config_option = click.option(
    "-c",
    "--config-name",
    default="maestral",
    type=ConfigName(),
    is_eager=True,
    expose_value=True,
    help="Run command with the given configuration.",
)


@main.command(section="Core Commands", help="Start the sync daemon.")
@click.option(
    "--foreground",
    "-f",
    is_flag=True,
    default=False,
    help="Starts Maestral in the foreground.",
)
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Print log messages to stdout."
)
@config_option
@catch_maestral_errors
def start(foreground: bool, verbose: bool, config_name: str) -> None:

    # ---- run setup if necessary ------------------------------------------------------

    # We run the setup in the current process. This avoids starting a subprocess despite
    # running with the --foreground flag, prevents leaving a zombie process if the setup
    # fails with an exception and does not confuse systemd.

    from .main import Maestral
    from .daemon import (
        start_maestral_daemon,
        start_maestral_daemon_process,
        is_running,
        Start,
    )

    check_for_updates()

    if is_running(config_name):
        click.echo("Daemon is already running.")
        return

    m = Maestral(config_name, log_to_stdout=verbose)

    if m.pending_link:  # this may raise KeyringAccessError
        link_dialog(m)

    if m.pending_dropbox_folder:
        path = select_dbx_path_dialog(config_name, allow_merge=True)

        while True:
            try:
                m.create_dropbox_directory(path)
                break
            except OSError:
                cli.warn(
                    "Could not create folder. Please make sure that you have "
                    "permissions to write to the selected location or choose a "
                    "different location."
                )

        include_all = cli.confirm("Would you like sync all folders?")

        if not include_all:
            # get all top-level Dropbox folders
            cli.info("Loading...")
            entries = m.list_folder("/", recursive=False)

            names = [
                cast(str, e["name"]) for e in entries if e["type"] == "FolderMetadata"
            ]

            choices = cli.select_multiple(
                "Choose which folders to include", options=names
            )

            excluded_paths = [
                f"/{name}" for index, name in enumerate(names) if index not in choices
            ]

            m.excluded_items = excluded_paths

    # free resources
    del m

    if foreground:
        # start our current process
        start_maestral_daemon(config_name, log_to_stdout=verbose, start_sync=True)
    else:

        # start daemon process
        cli.echo("Starting Maestral...", nl=False)

        res = start_maestral_daemon_process(
            config_name, log_to_stdout=verbose, start_sync=True
        )

        if res == Start.Ok:
            cli.echo(" " * 8 + OK)
        elif res == Start.AlreadyRunning:
            cli.echo(" " * 8 + "Already running.")
        else:
            cli.echo(" " * 8 + FAILED)
            cli.echo("Please check logs for more information.")


@main.command(section="Core Commands", help="Stop the sync daemon.")
@existing_config_option
def stop(config_name: str) -> None:
    stop_daemon_with_cli_feedback(config_name)


@main.command(section="Core Commands", help="Restart the sync daemon.")
@click.option(
    "--foreground",
    "-f",
    is_flag=True,
    default=False,
    help="Start the sync daemon in the foreground.",
)
@click.option(
    "--verbose", "-v", is_flag=True, default=False, help="Print log messages to stdout."
)
@existing_config_option
@click.pass_context
def restart(ctx, foreground: bool, verbose: bool, config_name: str) -> None:
    stop_daemon_with_cli_feedback(config_name)
    ctx.forward(start)


@main.command(section="Core Commands", help="Run the GUI if installed.")
@config_option
def gui(config_name: str) -> None:

    from packaging.version import Version
    from packaging.requirements import Requirement

    try:
        from importlib.metadata import entry_points, requires, version  # type: ignore
    except ImportError:
        from importlib_metadata import entry_points, requires, version  # type: ignore

    # find all "maestral_gui" entry points registered by other packages
    gui_entry_points = entry_points().get("maestral_gui")

    if not gui_entry_points or len(gui_entry_points) == 0:
        raise cli.CliException(
            "No maestral GUI installed. Please run 'pip3 install maestral[gui]'."
        )

    # check if 1st party defaults "maestral_cocoa" or "maestral_qt" are installed
    default_gui = "maestral_cocoa" if sys.platform == "darwin" else "maestral_qt"
    default_entry_point = next(
        (e for e in gui_entry_points if e.name == default_gui), None
    )

    if default_entry_point:
        # check gui requirements
        requirements = [Requirement(r) for r in requires("maestral")]  # type: ignore

        for r in requirements:
            if r.marker and r.marker.evaluate({"extra": "gui"}):
                version_str = version(r.name)
                if not r.specifier.contains(Version(version_str), prereleases=True):
                    raise cli.CliException(
                        f"{r.name}{r.specifier} required but you have {version_str}"
                    )

        # load entry point
        run = default_entry_point.load()

    else:
        # load any 3rd party GUI
        fallback_entry_point = next(iter(gui_entry_points))
        run = fallback_entry_point.load()

    run(config_name)


@main.command(section="Core Commands", help="Pause syncing.")
@existing_config_option
def pause(config_name: str) -> None:

    from .daemon import MaestralProxy

    try:
        with MaestralProxy(config_name) as m:
            m.pause_sync()
        cli.ok("Syncing paused.")
    except Pyro5.errors.CommunicationError:
        cli.echo("Maestral daemon is not running.")


@main.command(section="Core Commands", help="Resume syncing.")
@existing_config_option
def resume(config_name: str) -> None:
    from .daemon import MaestralProxy

    try:
        with MaestralProxy(config_name) as m:
            if not check_for_fatal_errors(m):
                m.resume_sync()
                cli.ok("Syncing resumed.")

    except Pyro5.errors.CommunicationError:
        cli.echo("Maestral daemon is not running.")


@main.command(section="Core Commands", help="Link with a Dropbox account.")
@click.option(
    "-r",
    "relink",
    is_flag=True,
    default=False,
    help="Relink to the current account. Keeps the sync state.",
)
@config_option
@catch_maestral_errors
def link(relink: bool, config_name: str) -> None:
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        if m.pending_link or relink:
            link_dialog(m)
        else:
            cli.echo(
                "Maestral is already linked. Use the option "
                "'-r' to relink to the same account."
            )


@main.command(
    section="Core Commands",
    help="""
Unlinks your Dropbox account.

If Maestral is running, it will be stopped before unlinking.
""",
)
@existing_config_option
@catch_maestral_errors
def unlink(config_name: str) -> None:

    if cli.confirm("Are you sure you want unlink your account?", default=False):

        from .main import Maestral

        stop_daemon_with_cli_feedback(config_name)
        m = Maestral(config_name)
        m.unlink()

        cli.ok("Unlinked Maestral.")


@main.command(section="Information", help="Show the status of the daemon.")
@existing_config_option
@catch_maestral_errors
def status(config_name: str) -> None:

    from .daemon import MaestralProxy

    check_for_updates()

    try:
        with MaestralProxy(config_name) as m:

            n_errors = len(m.sync_errors)
            color = "red" if n_errors > 0 else "green"
            n_errors_str = click.style(str(n_errors), fg=color)
            cli.echo("")
            cli.echo("Account:      {}".format(m.get_state("account", "email")))
            cli.echo("Usage:        {}".format(m.get_state("account", "usage")))
            cli.echo("Status:       {}".format(m.status))
            cli.echo("Sync threads: {}".format("Running" if m.running else "Stopped"))
            cli.echo("Sync errors:  {}".format(n_errors_str))
            cli.echo("")

            check_for_fatal_errors(m)

            sync_errors = m.sync_errors

            if len(sync_errors) > 0:

                path_column = cli.Column(title="Path")
                message_column = cli.Column(title="Error", wraps=True)

                for error in sync_errors:
                    path_column.append(error["dbx_path"])
                    message_column.append("{title}. {message}".format(**error))

                table = cli.Table([path_column, message_column])

                table.echo()
                cli.echo("")

    except Pyro5.errors.CommunicationError:
        cli.echo("Maestral daemon is not running.")


@main.command(
    section="Information",
    help="""
Show the sync status of a local file or folder.

Returned value will be 'uploading', 'downloading', 'up to date', 'error', or 'unwatched'
(for files outside of the Dropbox directory). This will always be 'unwatched' if syncing
is paused. This command can be used to for instance to query information for a plugin to
a file-manager.
""",
)
@click.argument("local_path", type=click.Path(exists=True, resolve_path=True))
@existing_config_option
def file_status(local_path: str, config_name: str) -> None:

    from .daemon import MaestralProxy

    try:
        with MaestralProxy(config_name) as m:
            stat = m.get_file_status(local_path)
            cli.echo(stat)

    except Pyro5.errors.CommunicationError:
        cli.echo("unwatched")


@main.command(section="Information", help="Live view of all items being synced.")
@existing_config_option
@catch_maestral_errors
def activity(config_name: str) -> None:

    import curses
    from .utils import natural_size
    from .daemon import MaestralProxy

    try:
        with MaestralProxy(config_name) as m:

            if check_for_fatal_errors(m):
                return

            def curses_loop(screen) -> None:  # no type hints for screen provided yet

                curses.use_default_colors()  # don't change terminal background
                screen.nodelay(1)  # sets `screen.getch()` to non-blocking

                while True:

                    height, width = screen.getmaxyx()

                    # create header
                    lines = [f"Status: {m.status}, Sync errors: {len(m.sync_errors)}"]
                    lines.append("")

                    # create table
                    filenames = ["Path"]
                    states = ["Status"]
                    col_len = 4

                    for event in m.get_activity(limit=height - 3):

                        dbx_path = cast(str, event["dbx_path"])
                        direction = cast(str, event["direction"])
                        state = cast(str, event["status"])
                        size = cast(int, event["size"])
                        completed = cast(int, event["completed"])

                        filename = os.path.basename(dbx_path)
                        filenames.append(filename)

                        arrow = "↓" if direction == "down" else "↑"

                        if completed > 0:
                            done_str = natural_size(completed, sep=False)
                            todo_str = natural_size(size, sep=False)
                            states.append(f"{done_str}/{todo_str} {arrow}")
                        else:
                            if state == "syncing" and direction == "up":
                                states.append("uploading")
                            elif state == "syncing" and direction == "down":
                                states.append("downloading")
                            else:
                                states.append(state)

                        col_len = max(len(filename), col_len)

                    for name, state in zip(filenames, states):  # create rows
                        lines.append(name.ljust(col_len + 2) + state)

                    # print to console screen
                    screen.clear()
                    try:
                        screen.addstr("\n".join(lines))
                    except curses.error:
                        pass
                    screen.refresh()

                    # abort when user presses 'q', refresh otherwise
                    key = screen.getch()
                    if key == ord("q"):
                        break
                    elif key < 0:
                        time.sleep(1)

            # enter curses event loop
            curses.wrapper(curses_loop)

    except Pyro5.errors.CommunicationError:
        cli.echo("Maestral daemon is not running.")


@main.command(section="Information", help="Show recently changed or added files.")
@existing_config_option
def history(config_name: str) -> None:

    from datetime import datetime
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:
        history = m.get_history()

    table = cli.Table(
        [
            cli.Column("Path", elide=cli.Elide.Leading),
            cli.Column("Change"),
            cli.Column("Time"),
        ]
    )

    for event in history:

        dbx_path = cast(str, event["dbx_path"])
        change_type = cast(str, event["change_type"])
        change_time_or_sync_time = cast(float, event["change_time_or_sync_time"])
        dt = datetime.fromtimestamp(change_time_or_sync_time)

        table.append([dbx_path, change_type, dt])

    cli.echo("")
    table.echo()
    cli.echo("")


@main.command(section="Information", help="List contents of a Dropbox directory.")
@click.argument("dropbox_path", type=click.Path(), default="")
@click.option(
    "-l",
    "--long",
    is_flag=True,
    default=False,
    help="Show output in long format with metadata.",
)
@click.option(
    "-d",
    "--include-deleted",
    is_flag=True,
    default=False,
    help="Include deleted items in listing.",
)
@existing_config_option
@catch_maestral_errors
def ls(long: bool, dropbox_path: str, include_deleted: bool, config_name: str) -> None:

    from datetime import datetime
    from .utils import natural_size
    from .daemon import MaestralProxy

    if not dropbox_path.startswith("/"):
        dropbox_path = "/" + dropbox_path

    with MaestralProxy(config_name, fallback=True) as m:

        cli.echo("Loading...\r", nl=False)

        entries = m.list_folder(
            dropbox_path,
            recursive=False,
            include_deleted=include_deleted,
        )
        entries.sort(key=lambda x: cast(str, x["name"]).lower())

        if long:

            to_short_type = {
                "FileMetadata": "file",
                "FolderMetadata": "folder",
                "DeletedMetadata": "deleted",
            }

            table = cli.Table(
                columns=[
                    cli.Column("Name"),
                    cli.Column("Type"),
                    cli.Column("Size", align=cli.Align.Right),
                    cli.Column("Shared"),
                    cli.Column("Syncing"),
                    cli.Column("Last Modified"),
                ]
            )

            for entry in entries:

                item_type = to_short_type[cast(str, entry["type"])]
                name = cast(str, entry["name"])
                path_lower = cast(str, entry["path_lower"])

                text = "shared" if "sharing_info" in entry else "private"
                color = "bright_black" if text == "private" else None
                shared_field = cli.TextField(text, fg=color)

                excluded_status = m.excluded_status(path_lower)
                color = "green" if excluded_status == "included" else None
                text = "✓" if excluded_status == "included" else excluded_status
                excluded_field = cli.TextField(text, fg=color)

                if "size" in entry:
                    size = natural_size(cast(float, entry["size"]))
                else:
                    size = "-"

                dt_field: cli.Field

                if "client_modified" in entry:
                    cm = cast(str, entry["client_modified"])
                    dt = datetime.strptime(cm, "%Y-%m-%dT%H:%M:%S%z").astimezone()
                    dt_field = cli.DateField(dt)
                else:
                    dt_field = cli.TextField("-")

                table.append(
                    [name, item_type, size, shared_field, excluded_field, dt_field]
                )

            cli.echo(" " * 15)
            table.echo()
            cli.echo(" " * 15)

        else:

            grid = cli.Grid()

            for entry in entries:
                name = cast(str, entry["name"])
                color = "blue" if entry["type"] == "DeletedMetadata" else None

                grid.append(cli.TextField(name, fg=color))

            grid.echo()


@main.command(section="Information", help="Show linked Dropbox account information.")
@existing_config_option
def account_info(config_name: str) -> None:
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        email = m.get_state("account", "email")
        account_type = m.get_state("account", "type").capitalize()
        usage = m.get_state("account", "usage")
        dbid = m.get_conf("account", "account_id")

    cli.echo("")
    cli.echo(f"Email:         {email}")
    cli.echo(f"Account-type:  {account_type}")
    cli.echo(f"Usage:         {usage}")
    cli.echo(f"Dropbox-ID:    {dbid}")
    cli.echo("")


@main.command(section="Information", help="List all configured Dropbox accounts.")
def configs() -> None:

    from .daemon import is_running
    from .config import (
        MaestralConfig,
        MaestralState,
        list_configs,
        remove_configuration,
    )

    # clean up stale configs
    config_names = list_configs()

    for name in config_names:
        dbid = MaestralConfig(name).get("account", "account_id")
        if dbid == "" and not is_running(name):
            remove_configuration(name)

    # display remaining configs
    names = list_configs()
    emails = [MaestralState(c).get("account", "email") for c in names]

    table = cli.Table([cli.Column("Config name", names), cli.Column("Account", emails)])

    cli.echo("")
    table.echo()
    cli.echo("")


@main.command(
    section="Information", help="Return the version number and other information."
)
def about() -> None:

    year = time.localtime().tm_year

    cli.echo("")
    cli.echo(f"Version:    {__version__}")
    cli.echo(f"Website:    {__url__}")
    cli.echo(f"Copyright:  (c) 2018-{year}, {__author__}.")
    cli.echo("")


# ======================================================================================
# Settings
# ======================================================================================


@main.command(
    section="Settings",
    help="""
Automatically start the sync daemon on login.

A systemd or launchd service will be created to start a sync daemon for the given
configuration on user login.
""",
)
@click.option("--yes", "-Y", is_flag=True, default=False)
@click.option("--no", "-N", is_flag=True, default=False)
@existing_config_option
def autostart(yes: bool, no: bool, config_name: str) -> None:

    from .autostart import AutoStart

    auto_start = AutoStart(config_name)

    if not auto_start.implementation:
        cli.echo(
            "Autostart is currently not supported for your platform.\n"
            "Autostart requires systemd on Linux or launchd on macOS."
        )
        return

    if yes or no:
        if yes:
            auto_start.enable()
            cli.ok("Enabled start on login.")
        else:
            auto_start.disable()
            cli.ok("Disabled start on login.")
    else:
        if auto_start.enabled:
            cli.echo("Autostart is enabled. Use -N to disable.")
        else:
            cli.echo("Autostart is disabled. Use -Y to enable.")


@main.group(section="Settings", help="View and manage excluded folders.")
def excluded():
    pass


@excluded.command(name="list", help="List all excluded files and folders.")
@existing_config_option
def excluded_list(config_name: str) -> None:
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        excluded_items = m.excluded_items
        excluded_items.sort()

        if len(excluded_items) == 0:
            cli.echo("No excluded files or folders.")
        else:
            for item in excluded_items:
                cli.echo(item)


@excluded.command(
    name="add",
    help="Add a file or folder to the excluded list and re-sync.",
)
@click.argument("dropbox_path", type=click.Path())
@existing_config_option
@catch_maestral_errors
def excluded_add(dropbox_path: str, config_name: str) -> None:
    from .daemon import MaestralProxy

    if not dropbox_path.startswith("/"):
        dropbox_path = "/" + dropbox_path

    if dropbox_path == "/":
        raise cli.CliException("Cannot exclude the root directory.")

    with MaestralProxy(config_name, fallback=True) as m:
        m.exclude_item(dropbox_path)
        cli.ok(f"Excluded '{dropbox_path}'.")


@excluded.command(
    name="remove",
    help="""
Remove a file or folder from the excluded list and re-sync.

It is safe to call this method with items which have already been included, they will
not be downloaded again. If the given path lies inside an excluded folder, the parent
folder will be included as well (but no other items inside it).
""",
)
@click.argument("dropbox_path", type=click.Path())
@existing_config_option
@catch_maestral_errors
def excluded_remove(dropbox_path: str, config_name: str) -> None:
    from .daemon import MaestralProxy

    if not dropbox_path.startswith("/"):
        dropbox_path = "/" + dropbox_path

    if dropbox_path == "/":
        return cli.echo("The root directory is always included")

    try:
        with MaestralProxy(config_name) as m:
            m.include_item(dropbox_path)
            cli.ok(f"Included '{dropbox_path}'. Now downloading...")

    except Pyro5.errors.CommunicationError:
        raise cli.CliException("Daemon must be running to download folders.")


@main.group(section="Settings", help="Manage desktop notifications.")
def notify():
    pass


@notify.command(
    name="level",
    help="Get or set the level for desktop notifications.",
)
@click.argument(
    "level_name",
    required=False,
    type=click.Choice(["ERROR", "SYNCISSUE", "FILECHANGE"]),
)
@existing_config_option
def notify_level(level_name: str, config_name: str) -> None:

    from .notify import MaestralDesktopNotifier as Notifier
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:
        if level_name:
            m.notification_level = Notifier.level_name_to_number(level_name)
            cli.ok(f"Notification level set to {level_name}.")
        else:
            level_name = Notifier.level_number_to_name(m.notification_level)
            cli.echo(f"Notification level: {level_name}.")


@notify.command(
    name="snooze",
    help="Snooze desktop notifications of file changes.",
)
@click.argument("minutes", type=click.IntRange(min=0))
@existing_config_option
def notify_snooze(minutes: int, config_name: str) -> None:
    from .daemon import MaestralProxy

    try:
        with MaestralProxy(config_name) as m:
            m.notification_snooze = minutes
    except Pyro5.errors.CommunicationError:
        cli.echo("Maestral daemon is not running.")
    else:
        if minutes > 0:
            cli.ok(
                f"Notifications snoozed for {minutes} min. " "Set snooze to 0 to reset."
            )
        else:
            cli.ok("Notifications enabled.")


@main.command(
    section="Settings",
    help="""
Enable or disables sharing of error reports.

Sharing is disabled by default. If enabled, error reports are shared with bugsnag and no
personal information will typically be collected. Shared tracebacks may however include
file names, depending on the error.
""",
)
@click.option("--yes", "-Y", is_flag=True, default=False)
@click.option("--no", "-N", is_flag=True, default=False)
@existing_config_option
def analytics(yes: bool, no: bool, config_name: str) -> None:
    from .daemon import MaestralProxy

    if yes or no:
        with MaestralProxy(config_name, fallback=True) as m:
            m.analytics = yes

        status_str = "Enabled" if yes else "Disabled"
        cli.ok(f"{status_str} automatic error reports.")
    else:
        with MaestralProxy(config_name, fallback=True) as m:
            enabled = m.analytics

        if enabled:
            cli.echo("Analytics are enabled. Use -N to disable")
        else:
            cli.echo("Analytics are disabled. Use -Y to enable")


# ======================================================================================
# Maintenance
# ======================================================================================


@main.command(section="Maintenance", help="Move the local Dropbox folder.")
@click.argument("new_path", required=False, type=click.Path(writable=True))
@existing_config_option
def move_dir(new_path: str, config_name: str) -> None:
    from .daemon import MaestralProxy

    new_path = new_path or select_dbx_path_dialog(config_name)

    with MaestralProxy(config_name, fallback=True) as m:
        m.move_dropbox_directory(new_path)

    cli.ok(f"Dropbox folder moved to {new_path}.")


@main.command(
    section="Maintenance",
    help="""
Rebuild the sync index.

Rebuilding may take several minutes, depending on the size of your Dropbox.
""",
)
@existing_config_option
@catch_maestral_errors
def rebuild_index(config_name: str) -> None:

    import textwrap
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:

        width, height = click.get_terminal_size()

        msg = textwrap.fill(
            "Rebuilding the index may take several minutes, depending on the size of "
            "your Dropbox. Any changes to local files will be synced once rebuilding "
            "has completed. If you stop the daemon during the process, rebuilding will "
            "start again on the next launch.\nIf the daemon is not currently running, "
            "a rebuild will be schedules for the next startup.",
            width=width,
        )

        cli.echo(msg + "\n")

        if cli.confirm("Do you want to continue?", default=False):

            m.rebuild_index()

            if m._is_fallback:
                cli.ok("Rebuilding now. Run 'maestral status' to view progress.")
            else:
                cli.ok("Daemon is not running. Rebuilding scheduled for next startup.")


@main.command(
    section="Maintenance",
    help="""
Restore a previous version of a file.

If no revision number is given, old revisions will be listed.
""",
)
@click.argument("dropbox_path", type=click.Path())
@click.option("-v", "--rev", help="Revision to restore", default="")
@existing_config_option
@catch_maestral_errors
def restore(dropbox_path: str, rev: str, config_name: str) -> None:
    from datetime import datetime
    from .daemon import MaestralProxy

    if not dropbox_path.startswith("/"):
        dropbox_path = "/" + dropbox_path

    with MaestralProxy(config_name, fallback=True) as m:

        if not rev:
            cli.echo("Loading...\r", nl=False)
            entries = m.list_revisions(dropbox_path)
            dates = []
            for entry in entries:
                cm = cast(str, entry["client_modified"])
                dt = datetime.strptime(cm, "%Y-%m-%dT%H:%M:%S%z").astimezone()
                field = cli.DateField(dt)
                dates.append(field.format(40)[0])

            index = cli.select(
                message="Select a version to restore:",
                options=dates,
                hint="(↓ to see more)" if len(entries) > 6 else "",
            )
            rev = cast(str, entries[index]["rev"])

        m.restore(dropbox_path, rev)

    cli.ok(f'Restored {rev} to "{dropbox_path}"')


@main.group(section="Maintenance", help="View and manage the log.")
def log():
    pass


@log.command(name="show", help="Print logs to the console.")
@click.option(
    "--external", "-e", is_flag=True, default=False, help="Open in external program."
)
@existing_config_option
def log_show(external: bool, config_name: str) -> None:

    from .utils.appdirs import get_log_path

    log_file = get_log_path("maestral", config_name + ".log")

    if external:
        res = click.launch(log_file)
    else:
        try:
            with open(log_file) as f:
                text = f.read()
            click.echo_via_pager(text)
        except OSError:
            res = 1
        else:
            res = 0

    if res > 0:
        raise cli.CliException(f"Could not open log file at '{log_file}'")


@log.command(name="clear", help="Clear the log files.")
@existing_config_option
def log_clear(config_name: str) -> None:

    from .utils.appdirs import get_log_path

    log_dir = get_log_path("maestral")
    log_name = config_name + ".log"

    log_files = []

    for file_name in os.listdir(log_dir):
        if file_name.startswith(log_name):
            log_files.append(os.path.join(log_dir, file_name))

    try:
        for file in log_files:
            open(file, "w").close()
        cli.ok("Cleared log files.")
    except FileNotFoundError:
        cli.ok("Cleared log files.")
    except OSError:
        raise cli.CliException(
            f"Could not clear log at '{log_dir}'. " f"Please try to delete it manually"
        )


@log.command(name="level", help="Get or set the log level.")
@click.argument(
    "level_name",
    required=False,
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR"]),
)
@existing_config_option
def log_level(level_name: str, config_name: str) -> None:

    import logging
    from .daemon import MaestralProxy

    with MaestralProxy(config_name, fallback=True) as m:
        if level_name:
            m.log_level = cast(int, getattr(logging, level_name))
            cli.ok(f"Log level set to {level_name}.")
        else:
            level_name = logging.getLevelName(m.log_level)
            cli.echo(f"Log level: {level_name}")
